import re
from functools import lru_cache
from pathlib import Path

from notary.models import licenses


def levenshtein(s1, s2):
    """Calculates the Levenshtein distance between two strings. Shamelessly stolen from
    https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance#Python
    """
    if len(s1) < len(s2):
        return levenshtein(s2, s1)

    # len(s1) >= len(s2)
    if len(s2) == 0:
        return len(s1)

    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[
                j + 1
            ] + 1  # j+1 instead of j since previous_row and current_row are one character longer
            deletions = current_row[j] + 1  # than s2
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]


def guess_license(name):
    probable = []
    likely = []
    for lic in licenses:
        if name.lower() in lic.name.lower() or name.lower() in lic.__name__.lower():
            probable.append(lic)
            continue

        likely.append(lic)

    if probable:
        return probable

    return likely


@lru_cache(maxsize=32)
def find_license_files():
    """Returns a list of :class:`Path <Path>` objects representing existing LICENSE files
    in the current directory.
    """
    rule = re.compile('(?i)license(\.[a-zA-Z]*)?')
    return [path for path in Path('.').glob('*') if path.is_file() and rule.match(path.name)]



from abc import ABC, abstractmethod

from notary import LICENSE_DIR


class License(ABC):
    """An abstract base class for all licenses.

    :param author: Name of the entity granting the license (if it needs one).
    :param year: The year to be written to the license (if it needs one).

    Any class that extends this one needs to provide the following params:
    :param name: The human-readable name of the license.
    :param path: Instance of :class:`pathlib.Path <pathlib.Path>` representing the location
    of the license file.

    Classes that allow setting a author or a year also additionally need to specify:
    :attr author_placeholder: What (exactly) in the license file should be replaced
    by :param:`author`
    :attr year_placeholder: What (exactly) in the license file should be replaced
    by :param:`year`
    """

    author_placeholder = None
    year_placeholder = None

    def __init__(self, author=None, year=None):
        self._author = author
        self._year = year

    @property
    @abstractmethod
    def name(self):
        pass

    @property
    @abstractmethod
    def path(self):
        pass

    @property
    def author(self):
        return self._author

    @property
    def year(self):
        return self._year

    @property
    def content(self):
        with self.path.open('r') as f:
            content = f.read()

        if self.author_placeholder:
            content = content.replace(self.author_placeholder, self.author, 1)

        if self.year_placeholder:
            content = content.replace(self.year_placeholder, str(self.year), 1)

        return content


class AGPL3(License):
    name = 'GNU Affero General Public License version 3'
    path = LICENSE_DIR.joinpath('agpl-3.0')


class Apache(License):
    name = 'Apache License 2.0'
    path = LICENSE_DIR.joinpath('apache-2.0.md')
    author_placeholder = '{name of copyright owner}'
    year_placeholder = '{yyyy}'


class GPL3(License):
    name = 'GNU General Public License v3.0'
    path = LICENSE_DIR.joinpath('gpl-3.0.md')


class LGPL3(License):
    name = 'GNU Lesser General Public License v3.0'
    path = LICENSE_DIR.joinpath('lgpl-3.0.md')


class MIT(License):
    name = 'MIT License'
    path = LICENSE_DIR.joinpath('mit.md')
    author_placeholder = '[fullname]'
    year_placeholder = '[year]'


class MPL(License):
    name = 'Mozilla Public License Version 2.0'
    path = LICENSE_DIR.joinpath('mpl-2.0.md')


class Unlicense(License):
    name = 'The Unlicense'
    path = LICENSE_DIR.joinpath('unlicense.md')


licenses = [
    AGPL3,
    Apache,
    GPL3,
    LGPL3,
    MIT,
    MPL,
    Unlicense,
]
